# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
from contextlib import contextmanager

from scipy.sparse import load_npz

from earthkit.regrid.gridspec import GridSpec
from earthkit.regrid.utils import no_progress_bar
from earthkit.regrid.utils.download import download_and_cache

LOG = logging.getLogger(__name__)

_URL = "https://get.ecmwf.int/repository/earthkit/regrid/matrices"
_INDEX_FILENAME = "index.json"


class UrlAccessor:
    def __init__(self, url):
        self.url = url

    def index_path(self):
        # checking the out of date status does not work for this file,
        # so we have to force the download using Force=True
        try:
            url = os.path.join(self.url, _INDEX_FILENAME)
            path = download_and_cache(
                url,
                owner="url",
                verify=True,
                force=True,
                chunk_size=1024 * 1024,
                http_headers=None,
                update_if_out_of_date=True,
                progress_bar=no_progress_bar,
                maximum_retries=5,
                retry_after=10,
            )
        except Exception:
            LOG.error(f"Could not download index file={url}")
            raise

        return path

    def matrix_path(self, name):
        try:
            url = os.path.join(self.url, name)
            path = download_and_cache(
                url,
                owner="url",
                verify=True,
                force=None,
                chunk_size=1024 * 1024,
                http_headers=None,
                update_if_out_of_date=False,
                maximum_retries=5,
                retry_after=10,
            )
        except Exception:
            LOG.error(f"Could not download matrix file={url}")
            raise

        return path


class LocalAccessor:
    """Only used for test purposes"""

    def __init__(self, path):
        self.path = path

    def index_path(self):
        return os.path.join(self.path, _INDEX_FILENAME)

    def matrix_path(self, name):
        return os.path.join(self.path, name)


@contextmanager
def _use_local_index(path):
    """Context manager for testing only. Allow using local index
    file and matrices.
    """
    DB.clear_index()
    DB.accessor = LocalAccessor(path)
    try:
        yield None
    finally:
        DB.clear_index()
        DB.accessor = UrlAccessor(_URL)


class MatrixDb:
    def __init__(self):
        self._index = None
        self.accessor = UrlAccessor(_URL)

    @property
    def index(self):
        if self._index is None:
            self.load_index()
        return self._index

    def load_index(self):
        self._index = {}
        path = self.accessor.index_path()

        with open(path, "r") as f:
            index = json.load(f)
            for name, entry in index.items():
                # it is possible that the inventory is already updated with new
                # a gridspecs type, but a given earthkit-regrid version is not
                # yet supporting it. In this case loading the index should not crash.
                try:
                    in_gs = GridSpec.from_dict(entry["input"])
                    out_gs = GridSpec.from_dict(entry["output"])
                    entry["input"] = in_gs
                    entry["output"] = out_gs
                    # print(f"{in_gs=}")
                    # print(f"{out_gs=}")
                    self._index[name] = entry
                # when the inventory is already updated with new a gridspecs type
                # but the old code is not yet supporting it should not crash
                except Exception:
                    pass

    def clear_index(self):
        """For testing only"""
        self._index = None

    def find(self, gridspec_in, gridspec_out, matrix_version=None):
        entry = self.find_entry(gridspec_in, gridspec_out)

        if entry is not None:
            versions = entry["versions"]
            if matrix_version is not None:
                if matrix_version not in versions:
                    raise ValueError(f"Unsupported matrix_version={matrix_version}")
            else:
                matrix_version = sorted(versions)[0]

            z = self.load_matrix(entry["name"], matrix_version)
            return z, entry["output"]["shape"]
        return None, None

    def find_entry(self, gridspec_in, gridspec_out):
        gridspec_in = GridSpec.from_dict(gridspec_in)
        gridspec_out = GridSpec.from_dict(gridspec_out)

        if gridspec_in is None or gridspec_out is None:
            return None

        for _, entry in self.index.items():
            if gridspec_in == entry["input"] and gridspec_out == entry["output"]:
                return entry

        return None

    def load_matrix(self, name, version):
        name = f"{name}-{version}.npz"
        path = self.accessor.matrix_path(name)
        z = load_npz(path)
        return z


DB = MatrixDb()



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from utils.grid import make_grid_id
from utils.matrix import make_matrix

LOG = logging.getLogger(__name__)


# shorthand for grid definition:
# ll : 1 - > 1/1
# O_rgg: 32 -> O32
# N_rgg: 32 -> N32
# H_ring: 32 -> H32_ring
# H_nested: 32 -> H32_nested


logging.basicConfig(level=logging.DEBUG)

in_grids = {
    "ll": [],
    "O_rgg": [],
    "N_rgg": [],
    # "H_ring": [1024, 512, 256, 128, 64, 32, 16, 8, 4, 2],
    "H_nested": [1024, 512, 256, 128, 64, 32, 16, 8, 4, 2],
}
out_grids = {
    "ll": [
        0.1,
        0.15,
        0.2,
        0.25,
        0.3,
        0.4,
        0.5,
        0.6,
        0.7,
        0.75,
        0.8,
        0.9,
        1,
        1.2,
        1.25,
        1.4,
        1.5,
        1.6,
        1.8,
        2,
        2.5,
        5,
        10,
    ]
}

matrix_dir = "matrices_29022024"
# index_file = "index/index.json"

in_grids = make_grid_id(in_grids)
out_grids = make_grid_id(out_grids)

LOG.debug(f"{in_grids=}")
LOG.debug(f"{out_grids=}")

for g_in in in_grids:
    for g_out in out_grids:
        if g_in != g_out:
            make_matrix(g_in, g_out, matrix_dir, index_file=None, download_index=False)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os

import numpy as np
import pytest

from earthkit.regrid import interpolate
from earthkit.regrid.db import DB, _use_local_index

PATH = os.path.join(os.path.dirname(__file__), "data", "local")

MATRIX_VERSION = "011801"


def file_in_testdir(filename):
    return os.path.join(PATH, filename)


def test_local_index():
    with _use_local_index(PATH):
        # we have an extra unsupported entry in the index file. We have
        # be sure the DB is loaded correctly bypassing the the unsupported
        # entry.
        import json

        index_path = DB.accessor.index_path()
        with open(index_path, "r") as f:
            assert len(json.load(f)) == 5

        assert len(DB.index) == 4

        r = DB.find_entry({"grid": [5, 5]}, {"grid": [10, 10]})
        assert r

        r = DB.find_entry({"grid": "O32"}, {"grid": [10, 10]})
        assert r

        r = DB.find_entry({"grid": "N32"}, {"grid": [10, 10]})
        assert r

        r = DB.find_entry({"grid": "O64"}, {"grid": [10, 10]})
        assert r is None


@pytest.mark.parametrize(
    "_kwargs", [({"matrix_version": MATRIX_VERSION}), ({"matrix_version": None}), ({})]
)
def test_local_ll_to_ll(_kwargs):
    with _use_local_index(PATH):
        v_in = np.load(file_in_testdir("in_5x5.npz"))["arr_0"]
        v_ref = np.load(file_in_testdir(f"out_5x5_10x10_{MATRIX_VERSION}.npz"))["arr_0"]
        v_res = interpolate(v_in, {"grid": [5, 5]}, {"grid": [10, 10]}, **_kwargs)

        assert v_res.shape == (19, 36)
        assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize(
    "_kwargs", [({"matrix_version": MATRIX_VERSION}), ({"matrix_version": None}), ({})]
)
def test_local_gg_to_ll_1(_kwargs):
    with _use_local_index(PATH):
        v_in = np.load(file_in_testdir("in_O32.npz"))["arr_0"]
        v_ref = np.load(file_in_testdir(f"out_O32_10x10_{MATRIX_VERSION}.npz"))["arr_0"]
        v_res = interpolate(v_in, {"grid": "O32"}, {"grid": [10, 10]}, **_kwargs)

        assert v_res.shape == (19, 36)
        assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize(
    "_kwargs", [({"matrix_version": MATRIX_VERSION}), ({"matrix_version": None}), ({})]
)
def test_local_gg_to_ll_2(_kwargs):
    with _use_local_index(PATH):
        v_in = np.load(file_in_testdir("in_N32.npz"))["arr_0"]
        v_ref = np.load(file_in_testdir(f"out_N32_10x10_{MATRIX_VERSION}.npz"))["arr_0"]
        v_res = interpolate(v_in, {"grid": "N32"}, {"grid": [10, 10]}, **_kwargs)

        assert v_res.shape == (19, 36)
        assert np.allclose(v_res.flatten(), v_ref)


@pytest.mark.parametrize(
    "gs_in, gs_out",
    [
        ({"grid": [5, 5]}, {"grid": [10, 10]}),
        ({"grid": [5, 5], "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": [5, 5], "area": [90, 0, -90, 355]}, {"grid": [10, 10]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "area": [90, 0, -90, 360]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "area": [90, 0, -90, 350]}),
        (
            {"grid": [5, 5], "area": [90, 0, -90, 360]},
            {"grid": [10, 10], "area": [90, 0, -90, 360]},
        ),
        ({"grid": [5, 5], "shape": [37, 72]}, {"grid": [10, 10], "shape": [19, 36]}),
        ({"grid": [5, 5], "shape": [37, 72]}, {"grid": [10, 10]}),
        ({"grid": [5, 5]}, {"grid": [10, 10], "shape": [19, 36]}),
        ({"grid": "O32"}, {"grid": [10, 10]}),
        ({"grid": "O32", "shape": [5248]}, {"grid": [10, 10]}),
        ({"grid": "O32", "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": "O32", "area": [87.8638, 0, -87.8638, 357.5]}, {"grid": [10, 10]}),
        ({"grid": "O32", "global": 1}, {"grid": [10, 10]}),
        (
            {"grid": "O32", "global": 1, "area": [87.8638, 0, -87.8638, 357.5]},
            {"grid": [10, 10]},
        ),
        ({"grid": "N32"}, {"grid": [10, 10]}),
        ({"grid": "N32", "shape": [6114]}, {"grid": [10, 10]}),
        ({"grid": "N32", "area": [90, 0, -90, 360]}, {"grid": [10, 10]}),
        ({"grid": "N32", "area": [87.8638, 0, -87.8638, 357.188]}, {"grid": [10, 10]}),
        ({"grid": "N32", "global": 1}, {"grid": [10, 10]}),
        (
            {"grid": "N32", "global": 1, "area": [87.8638, 0, -87.8638, 357.188]},
            {"grid": [10, 10]},
        ),
        ({"grid": "H128"}, {"grid": [1, 1]}),
        ({"grid": "H128", "ordering": "ring"}, {"grid": [1, 1]}),
    ],
)
def test_local_gridspec_ok(gs_in, gs_out):
    with _use_local_index(PATH):
        r = DB.find_entry(gs_in, gs_out)
        assert r, f"gs_in={gs_in} gs_out={gs_out}"


@pytest.mark.parametrize(
    "gs_in,gs_out,err",
    [
        ({"grid": [1, 1]}, {"grid": [2, 2]}, None),
        ({"grid": [5, 5], "area": [90, 0, -90, 350]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90.001, 0, -90, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0, -89.0001, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0.001, -90, 360]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": [5, 5], "area": [90, 10, -90, 370]}, {"grid": [10, 10]}, None),
        ({"grid": "G1280", "shape": 6599680}, {"grid": [10, 10]}, ValueError),
        ({"grid": "O32", "shape": 6599680}, {"grid": [10, 10]}, None),
        ({"grid": "O32", "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": "O32", "area": [90, -0.1, -90, 360]}, {"grid": [10, 10]}, None),
        (
            {"grid": "O32", "area": [87.8638, 0, -87.8638, 357.6]},
            {"grid": [10, 10]},
            None,
        ),
        (
            {"grid": "O32", "area": [87.8638, 0.01, -87.8638, 357.5]},
            {"grid": [10, 10]},
            None,
        ),
        ({"grid": "N32", "shape": 6599680}, {"grid": [10, 10]}, None),
        ({"grid": "N32", "area": [90, 0, -90, 359.999]}, {"grid": [10, 10]}, None),
        ({"grid": "N32", "area": [90, -0.1, -90, 360]}, {"grid": [10, 10]}, None),
        (
            {"grid": "N32", "area": [87.8638, 0, -87.8638, 357.189]},
            {"grid": [10, 10]},
            None,
        ),
        (
            {"grid": "N32", "area": [87.8638, 0.01, -87.8638, 357.188]},
            {"grid": [10, 10]},
            None,
        ),
        ({"grid": "H128", "ordering": "any"}, {"grid": [1, 1]}, ValueError),
    ],
)
def test_local_gridspec_bad(gs_in, gs_out, err):
    with _use_local_index(PATH):
        if err:
            with pytest.raises(err):
                r = DB.find_entry(gs_in, gs_out)
        else:
            r = DB.find_entry(gs_in, gs_out)
            assert r is None, f"gs_in={gs_in} gs_out={gs_out}"



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import os

from .mir import mir_cached_matrix_to_file


def regular_ll(entry):
    bb = entry["bbox"]
    d = {
        "grid": [entry["increments"][x] for x in ("west_east", "south_north")],
        "shape": [entry["nj"], entry["ni"]],
        "area": [bb["north"], bb["west"], bb["south"], bb["east"]],
    }
    if "global" in entry:
        d["global"] = entry["global"]
    return d


def reduced_gg(entry):
    pl = entry["pl"]
    G = "O" if pl[1] - pl[0] == 4 else "N"
    N = entry["N"]
    bb = entry["bbox"]

    d = {
        "grid": f"{G}{N}",
        "shape": [sum(pl)],
        "area": [bb["north"], bb["west"], bb["south"], bb["east"]],
    }

    if "global" in entry:
        d["global"] = entry["global"]

    return d


def healpix(entry):
    d = {"grid": entry["grid"], "ordering": entry["ordering"]}
    return d


def make_matrix(
    input_path,
    output_path,
    index_file=None,
    global_input=None,
    global_output=None,
    version=None,
):
    with open(input_path) as f:
        entry = json.load(f)

    cache_file = entry.pop("cache_file")
    name, _ = os.path.splitext(os.path.basename(cache_file))

    # print(f"MATRICES={MATRICES}")

    if version is None:
        version = "0" * 6
    else:
        version = "".join([f"{int(x):02d}" for x in version.split(".")])

    os.makedirs(output_path, exist_ok=True)

    print(f"entry={entry}")
    npz_file = os.path.join(output_path, f"{name}-{version}.npz")

    mir_cached_matrix_to_file(cache_file, npz_file)

    if index_file is None:
        index_file = os.path.join(output_path, "index.json")

    if os.path.exists(index_file):
        with open(index_file) as f:
            index = json.load(f)
    else:
        index = {}

    def convert(x):
        proc = globals()[x["type"]]
        return proc(x)

    if global_input is not None and "global" not in entry["input"]:
        entry["input"]["global"] = 1 if global_input else 0

    if global_output is not None and "global" not in entry["output"]:
        entry["output"]["global"] = 1 if global_output else 0

    index[name] = dict(
        name=name,
        versions=[version],
        input=convert(entry["input"]),
        output=convert(entry["output"]),
    )

    with open(index_file, "w") as f:
        json.dump(index, f, indent=4)

    print("Written", npz_file)
    print("Written", index_file)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.regrid.db import DB


def interpolate(values, source_gridspec, target_gridspec, matrix_version=None):
    z, shape = DB.find(source_gridspec, target_gridspec, matrix_version=matrix_version)

    if z is None:
        raise ValueError("No matrix found that matches the source and target gridspecs")

    # This should check for 1D (GG) and 2D (LL) matrices
    values = values.reshape(-1, 1)

    values = z @ values

    return values.reshape(shape)
